import copy
from ..utils.config import load_all_configs
from .route import route_vac
from ..custom_logging import log
from ruamel.yaml import YAML
from io import StringIO
# check it here:
# https://editor.swagger.io/
from functools import lru_cache
import uuid

try:
    from google.cloud import servicecontrol_v1
    from google.protobuf.timestamp_pb2 import Timestamp
except ImportError:
    servicecontrol_v1 = None

def validate_api_key(api_key: str, service_name: str) -> bool:
    """
    Validate an API key against the service name e.g. 'endpoints-xxxx.a.run.app'
    """
    if not servicecontrol_v1:
        raise ImportError("Cloud Endpoints API key validation is required, install via `pip install sunholo[gcp]`")

    return _validate_api_key_cached(api_key, service_name)


@lru_cache(maxsize=1024)
def _validate_api_key_cached(api_key: str, service_name: str) -> bool:
    client = servicecontrol_v1.ServiceControllerClient()
    # Create a timestamp for the current time
    start_time = Timestamp()
    start_time.GetCurrentTime()

    request = servicecontrol_v1.CheckRequest(
        service_name=service_name,
        operation=servicecontrol_v1.Operation(
            operation_name=service_name.upper(),
            operation_id=str(uuid.uuid4()),
            consumer_id=f'api_key:{api_key}',
            start_time=start_time,
        )
    )
    log.debug(f"API_KEY checkRequest {request}")

    response = client.check(request=request)

    # Check if there are any check errors
    if response.check_errors:
        log.debug(f"API_KEY validation failed with errors: {response.check_errors}")
        return False

    log.debug(f"API_KEY validation succeeded - {response}")
    return True

def config_to_swagger():
    """
    Load configuration files and generate a Swagger specification.

    This function loads the 'vacConfig' and 'agentConfig' configuration files,
    validates their presence, and then generates a Swagger specification
    based on these configurations.

    Returns:
        str: The generated Swagger specification in YAML format.

    Raises:
        ValueError: If 'vacConfig' or 'agentConfig' is not loaded.

    Example:
    ```python
        swagger_yaml = config_to_swagger()
        print(swagger_yaml)
    ```
    """
    configs = load_all_configs()

    vac_config = configs.get('vacConfig')
    agent_config = configs.get('agentConfig')

    if not vac_config:
        raise ValueError("Need valid 'vacConfig' loaded")
    
    if not agent_config:
        raise ValueError("Need valid 'agentConfig' loaded")
    
    swag = generate_swagger(vac_config, agent_config)

    return swag

def generate_swagger(vac_config, agent_config):
    """
    Generate a Swagger specification based on the provided configurations.

    This function creates a Swagger specification using the provided 'vacConfig'
    and 'agentConfig'. It dynamically builds paths and responses based on the
    configurations.

    Args:
        vac_config (dict): The VAC configuration.
        agent_config (dict): The agent configuration.

    Returns:
        str: The generated Swagger specification in YAML format.

    Example:
    ```python
        vac_config = {
            'vac': {
                'service1': {
                    'llm': 'vertex',
                    'model': 'gemini-1.5-flash-001',
                    'agent': 'langserve'
                },
                'service2': {
                    'llm': 'openai',
                    'agent': 'crewai',
                    'secrets': ['OPENAI_API_KEY']
                }
            }
        }
        
        agent_config = {
            'agents': {
                'default': {
                    'stream': "{stem}/vac/streaming/{vector_name}",
                    'invoke': "{stem}/vac/{vector_name}",
                    'post': {
                        'stream': "{stem}/vac/streaming/{vector_name}",
                        'invoke': "{stem}/vac/{vector_name}",
                        'openai': "{stem}/openai/v1/chat/completions",
                        'openai-vac': "{stem}/openai/v1/chat/completions/{vector_name}"
                    },
                    'get': {
                        'home': "{stem}/",
                        'health': "{stem}/health"
                    },
                    'response': {
                        'invoke': {
                            '200': {
                                'description': 'Successful invocation response',
                                'schema': {
                                    'type': 'object',
                                    'properties': {
                                        'answer': {'type': 'string'},
                                        'source_documents': {
                                            'type': 'array',
                                            'items': {
                                                'type': 'object',
                                                'properties': {
                                                    'page_content': {'type': 'string'},
                                                    'metadata': {'type': 'string'}
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        
        swagger_yaml = generate_swagger(vac_config, agent_config)
        print(swagger_yaml)
        ```
    """
    swagger_template = {
        'swagger': '2.0',
        'info': {
            'title': 'Sunholo Multivac API - ${_BRANCH_NAME}',
            'description': """
Multivac (Multi-VAC) Cloud abstracts the interaction between the Cloud and GenAI applications via Virtual Agent Computers (VAC), allowing standardization of input and output to accelerate deployments. 
VACs support frameworks like Langchain, LlamaIndex, VertexAI, OpenAI, or custom libraries, operating within Docker containers.
Multivac features serverless architecture, robust security within a Virtual Private Cloud (VPC), and flexible user interfaces. 
Generate API keys (MULTIVAC_API_KEY) to enable access to private services.
Partners can deploy all services within their own cloud for complete ownership.
See more at https://dev.sunholo.com/
            """,
            'version': '0.1.0'
        },
        'host': '${_ENDPOINTS_HOST}',
        'basePath': '/v1',
        'schemes': ['https'],
        'produces': ['application/json'],
        'paths': {},
        'securityDefinitions': { 
            'ApiKeyAuth': {        # For private VAC endpoint
                'type': 'apiKey',
                'name': 'x-api-key', # Custom header name for API key
                'in': 'header'
            },
            'None': {              # For public documentation
                'type': 'apiKey',
                'name': 'allow',    # Dummy parameter for public access
                'in': 'query'       # Use query parameter to avoid interfering with other headers
            }
        }
    }
    vac_services = vac_config['vac']
    
    def configure_agent(vector_name, config, agent_config_paths):
        agent_type = config['agent']
        log.info(f'Configuring swagger for agent_type: {agent_type} for vector_name: {vector_name}')
        try:
            stem = route_vac(vector_name).strip()
        except ValueError:
            log.warning(f"Failed to find URL stem for {vector_name}/{agent_type} - skipping")
            return

        for method, endpoints in agent_config_paths.items():
            do_auth = True

            if method not in ['get', 'post', 'get-auth', 'post-noauth']:
                continue

            if method == 'get-auth':
                do_auth = True
                method = 'get'
            elif method == 'post-noauth':
                do_auth = False
                method = 'post'
            elif method == 'get':
                do_auth = False
            elif method == 'post':
                do_auth = True

            for endpoint_key, endpoint_template in endpoints.items():
                endpoint_template = endpoint_template.strip()
                endpoint_address = endpoint_template.replace("{stem}", stem).replace("{vector_name}", vector_name).strip()
                endpoint_path = endpoint_template.replace("{stem}", f"/{agent_type}").replace("{vector_name}", vector_name).strip()
                log.debug(f"Endpoint_template: {endpoint_template}")
                log.debug(f"endpoint address: {endpoint_address}")
                log.debug(f"endpoint_path: {endpoint_path}")
                if endpoint_path not in swagger_template['paths']:
                    swagger_template['paths'][endpoint_path] = {}

                operation_id = f"{method}_{agent_type}_{endpoint_key}_{vector_name}"

                security = [{'ApiKeyAuth': []}] if do_auth else [{'None': []}]

                swagger_template['paths'][endpoint_path][method] = {
                    'summary': f"{method.capitalize()} {vector_name}",
                    'operationId': operation_id,
                    'x-google-backend': {
                        'address': endpoint_address,
                        'protocol': 'h2',
                        'deadline':  180000  # Timeout of 3 minutes (1,800,00 milliseconds)
                    },
                    'security': security,
                    'responses': copy.deepcopy(agent_config_paths.get('response', {}).get(endpoint_key, {
                        '200': {
                            'description': 'Default - A successful response',
                            'schema': {
                                'type': 'string'
                            }
                        }
                    }))
                }

    for vector_name, config in vac_services.items():
        agent_type = config['agent']
        agent_config_paths = agent_config['agents'].get(agent_type, {})
        configure_agent(vector_name, config, agent_config_paths)

    default_agent_config = agent_config['agents'].get('default', {})

    for vector_name, config in vac_services.items():
        agent_type = config['agent']
        if agent_type in agent_config['agents']:
            continue
        log.info(f'Applying default configuration for agent_type: {agent_type} for vector_name: {vector_name}')
        configure_agent(vector_name, config, default_agent_config)

    yaml = YAML()
    yaml.width = 4096

    string_stream = StringIO()
    yaml.dump(swagger_template, string_stream)
    yaml_string = string_stream.getvalue()

    try:
        from openapi_spec_validator import validate
        swagger_dict = yaml.load(yaml_string)
        swagger_dict['host'] = 'api.sunholo.com'
        validated = validate(swagger_dict)
        log.info(f'swagger validation errors: {validated}')
    except ImportError:
        log.error("Must have openapi-spec-validator installed to validate - install via `pip install sunholo'[tools]'`")

    return yaml_string


import json
import traceback
import datetime
import uuid

from ...agents import extract_chat_history, handle_special_commands
from ...qna.parsers import parse_output
from ...streaming import start_streaming_chat
from ...archive import archive_qa
from ...custom_logging import log
from ...utils.config import load_config
from ...utils.version import sunholo_version
import os
from ...gcs.add_file import add_file_to_gcs, handle_base64_image
from ..swagger import validate_api_key
from datetime import datetime, timedelta

try:
    from flask import request, jsonify, Response
except ImportError:
    pass

try:
    from langfuse.decorators import langfuse_context, observe
except ImportError:
    pass    

# Cache dictionary to store validated API keys
api_key_cache = {}
cache_duration = timedelta(minutes=5)  # Cache duration

class VACRoutes:
    """
**Usage Example:**

```python
from agents.flask import VACRoutes

app = Flask(__name__)

def stream_interpreter(question, vector_name, chat_history, **kwargs):
    # Implement your streaming logic
    ...

def vac_interpreter(question, vector_name, chat_history, **kwargs):
    # Implement your static VAC logic
    ...

vac_routes = VACRoutes(app, stream_interpreter, vac_interpreter)

if __name__ == "__main__":
    app.run(debug=True)
```
    
    """
    def __init__(self, app, stream_interpreter, vac_interpreter):
        self.app = app
        self.stream_interpreter = stream_interpreter
        self.vac_interpreter = vac_interpreter
        self.register_routes()

    def register_routes(self):
        """
        Registers all the VAC routes for the Flask application.
        """
        # Basic routes
        self.app.route("/", methods=['GET'])(self.home)
        self.app.route("/health", methods=['GET'])(self.health)

        # Streaming VAC
        self.app.route('/vac/streaming/<vector_name>', methods=['POST'])(self.handle_stream_vac)

        # Static VAC
        self.app.route('/vac/<vector_name>', methods=['POST'])(self.handle_process_vac)

        # Authentication middleware
        self.app.before_request(self.check_authentication)

        # OpenAI health endpoint
        self.app.route('/openai/health', methods=['GET', 'POST'])(self.openai_health_endpoint)

        # OpenAI compatible endpoint
        self.app.route('/openai/v1/chat/completions', methods=['POST'])(self.handle_openai_compatible_endpoint)
        self.app.route('/openai/v1/chat/completions/<vector_name>', methods=['POST'])(self.handle_openai_compatible_endpoint)

    def home(self):
        return jsonify("OK")

    def health(self):
        return jsonify({"status": "healthy"})
    
    def make_openai_response(self, user_message, vector_name, answer):
        response_id = str(uuid.uuid4())
        log.info("openai response: Q: {user_message} to VECTOR_NAME: {vector_name} - A: {answer}")
        openai_response = {
            "id": response_id,
            "object": "chat.completion",
            "created": str(int(datetime.now().timestamp())),
            "model": vector_name,
            "system_fingerprint": sunholo_version(),
            "choices": [{
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": answer,
                },
                "logprobs": None,
                "finish_reason": "stop"
            }],
            "usage": {
                "prompt_tokens": len(user_message.split()),
                "completion_tokens": len(answer.split()),
                "total_tokens": len(user_message.split()) + len(answer.split())
            }
        }

        log.info(f"OpenAI response: {openai_response}")
        return jsonify(openai_response)


    def handle_stream_vac(self, vector_name):
        observed_stream_interpreter = observe()(self.stream_interpreter)
        prep = self.prep_vac(request, vector_name)
        log.debug(f"Processing prep: {prep}")
        trace = prep["trace"]
        span = prep["span"]
        command_response = prep["command_response"]
        vac_config = prep["vac_config"]
        all_input = prep["all_input"]

        if command_response:
            return jsonify(command_response)

        log.info(f'Streaming data with: {all_input}')
        if span:
            generation = span.generation(
                name="start_streaming_chat",
                metadata=vac_config,
                input = all_input,
                completion_start_time=datetime.datetime.now(),
                model=vac_config.get("model") or vac_config.get("llm")
            )

        def generate_response_content():

            for chunk in start_streaming_chat(question=all_input["user_input"],
                                                vector_name=vector_name,
                                                qna_func=observed_stream_interpreter,
                                                chat_history=all_input["chat_history"],
                                                wait_time=all_input["stream_wait_time"],
                                                timeout=all_input["stream_timeout"],
                                                #kwargs
                                                **all_input["kwargs"]
                                                ):
                if isinstance(chunk, dict) and 'answer' in chunk:
                    # When we encounter the dictionary, we yield it as a JSON string
                    # and stop the generator.
                    if trace:
                        chunk["trace"] = trace.id
                        chunk["trace_url"] = trace.get_trace_url()
                    archive_qa(chunk, vector_name)
                    if trace:
                        generation.end(output=json.dumps(chunk))
                        span.end(output=json.dumps(chunk))
                        trace.update(output=json.dumps(chunk))

                    return json.dumps(chunk)
            
                else:
                    # Otherwise, we yield the plain text chunks as they come in.
                    yield chunk
        
        # Here, the generator function will handle streaming the content to the client.
        response = Response(generate_response_content(), content_type='text/plain; charset=utf-8')
        response.headers['Transfer-Encoding'] = 'chunked'  

        log.debug(f"streaming response: {response}")
        if trace:
            generation.end(output=response)
            span.end(output=response)
            trace.update(output=response)

        return response

    def handle_process_vac(self, vector_name):
        observed_vac_interpreter = observe()(self.vac_interpreter)
        prep = self.prep_vac(request, vector_name)
        log.debug(f"Processing prep: {prep}")
        trace = prep["trace"]
        span = prep["span"]
        command_response = prep["command_response"]
        vac_config = prep["vac_config"]
        all_input = prep["all_input"]

        if command_response:
            return jsonify(command_response)

        try:
            if span:
                generation = span.generation(
                    name="vac_interpreter",
                    metadata=vac_config,
                    input = all_input,
                    model=vac_config.get("model") or vac_config.get("llm")
                )
            bot_output = observed_vac_interpreter(
                question=all_input["user_input"],
                vector_name=vector_name,
                chat_history=all_input["chat_history"],
                **all_input["kwargs"]
            )
            if span:
                generation.end(output=bot_output)
            # {"answer": "The answer", "source_documents": [{"page_content": "The page content", "metadata": "The metadata"}]}
            bot_output = parse_output(bot_output)
            if trace:
                bot_output["trace"] = trace.id
                bot_output["trace_url"] = trace.get_trace_url()
            archive_qa(bot_output, vector_name)
            log.info(f'==LLM Q:{all_input["user_input"]} - A:{bot_output}')


        except Exception as err: 
            bot_output = {'answer': f'QNA_ERROR: An error occurred while processing /vac/{vector_name}: {str(err)} traceback: {traceback.format_exc()}'}
    
        if trace:
            span.end(output=jsonify(bot_output))
            trace.update(output=jsonify(bot_output)) 

        # {'answer': 'output'}
        return jsonify(bot_output)

    def check_authentication(self):
        if request.path.startswith('/openai/'):
            log.debug(f'Request headers: {request.headers}')
            # the header forwarded
            auth_header = request.headers.get('X-Forwarded-Authorization')
            if auth_header:
                
                if auth_header.startswith('Bearer '):
                    api_key = auth_header.split(' ')[1]  # Assuming "Bearer <api_key>"
                else:
                    return jsonify({'error': 'Invalid authorization header does not start with "Bearer " - got: {auth_header}'}), 401
                
                endpoints_host = os.getenv('_ENDPOINTS_HOST')
                if not endpoints_host:
                    return jsonify({'error': '_ENDPOINTS_HOST environment variable not found'}), 401
                
                # Check cache first
                current_time = datetime.now()
                if api_key in api_key_cache:
                    cached_result, cache_time = api_key_cache[api_key]
                    if current_time - cache_time < cache_duration:
                        if not cached_result:
                            return jsonify({'error': 'Invalid cached API key'}), 401
                        else:
                            return  # Valid API key, continue to the endpoint
                    else:
                        # Cache expired, remove from cache
                        del api_key_cache[api_key]
                
                # Validate API key
                is_valid = validate_api_key(api_key, endpoints_host)
                # Update cache
                api_key_cache[api_key] = (is_valid, current_time)
                
                if not is_valid:
                    return jsonify({'error': 'Invalid API key'}), 401
            else:
                return jsonify({'error': 'Missing Authorization header'}), 401

    def openai_health_endpoint():
            return jsonify({'message': 'Success'})

    def handle_openai_compatible_endpoint(self, vector_name=None):
        data = request.get_json()
        log.info(f'openai_compatible_endpoint got data: {data} for vector: {vector_name}')

        vector_name = vector_name or data.pop('model', None)
        messages = data.pop('messages', None)
        chat_history = data.pop('chat_history', None)
        stream = data.pop('stream', False)

        if not messages:
            return jsonify({"error": "No messages provided"}), 400

        user_message = None
        image_uri = None
        mime_type = None

        for msg in reversed(messages):
            if msg['role'] == 'user':
                if isinstance(msg['content'], list):
                    for content_item in msg['content']:
                        if content_item['type'] == 'text':
                            user_message = content_item['text']
                        elif content_item['type'] == 'image_url':
                            base64_data = content_item['image_url']['url']
                            image_uri, mime_type = handle_base64_image(base64_data, vector_name)
                else:
                    user_message = msg['content']
                break

        if not user_message:
            return jsonify({"error": "No user message provided"}), 400
        else:
            log.info(f"User message: {user_message}")
    
        paired_messages = extract_chat_history(chat_history)
        command_response = handle_special_commands(user_message, vector_name, paired_messages)

        if command_response is not None:

            return self.make_openai_response(user_message, vector_name, command_response)
    
        if image_uri:
            data["image_uri"] = image_uri
            data["mime"] = mime_type

        all_input = {
            "user_input": user_message,
            "chat_history": chat_history,
            "kwargs": data
        }

        observed_stream_interpreter = observe()(self.stream_interpreter)

        response_id = str(uuid.uuid4())
        
        def generate_response_content():
            for chunk in start_streaming_chat(question=user_message,
                                            vector_name=vector_name,
                                            qna_func=observed_stream_interpreter,
                                            chat_history=all_input["chat_history"],
                                            wait_time=all_input.get("stream_wait_time", 1),
                                            timeout=all_input.get("stream_timeout", 60),
                                            **all_input["kwargs"]
                                            ):
                if isinstance(chunk, dict) and 'answer' in chunk:
                    openai_chunk = {
                        "id": response_id,
                        "object": "chat.completion.chunk",
                        "created": str(int(datetime.now().timestamp())),
                        "model": vector_name,
                        "system_fingerprint": sunholo_version(),
                        "choices": [{
                            "index": 0,
                            "delta": {"content": chunk['answer']},
                            "logprobs": None,
                            "finish_reason": None
                        }]
                    }
                    yield json.dumps(openai_chunk) + "\n"
                else:
                    log.info(f"Unknown chunk: {chunk}")

            final_chunk = {
                "id": response_id,
                "object": "chat.completion.chunk",
                "created": str(int(datetime.now().timestamp())),
                "model": vector_name,
                "system_fingerprint": sunholo_version(),
                "choices": [{
                    "index": 0,
                    "delta": {},
                    "logprobs": None,
                    "finish_reason": "stop"
                }]
            }
            yield json.dumps(final_chunk) + "\n"

        if stream:
            log.info("Streaming openai chunks")
            return Response(generate_response_content(), content_type='text/plain; charset=utf-8')

        try:
            observed_vac_interpreter = observe()(self.vac_interpreter)
            bot_output = observed_vac_interpreter(
                question=user_message,
                vector_name=vector_name,
                chat_history=all_input["chat_history"],
                **all_input["kwargs"]
            )
            bot_output = parse_output(bot_output)

            log.info(f"Bot output: {bot_output}")
            if bot_output:
                return self.make_openai_response(user_message, vector_name, bot_output.get('answer', ''))
            else:
                return self.make_openai_response(user_message, vector_name, 'ERROR: could not find an answer')

        except Exception as err:
            log.error(f"OpenAI response error: {str(err)} traceback: {traceback.format_exc()}")
        
            return self.make_openai_response(user_message, vector_name, f'ERROR: {str(err)}')


    def create_langfuse_trace(self, request, vector_name):
        try:
            from langfuse import Langfuse
            langfuse = Langfuse()
        except ImportError as err:
            print(f"No langfuse installed for agents.flask.register_qna_routes, install via `pip install sunholo[http]` - {str(err)}")
            
            return None

        user_id = request.headers.get("X-User-ID")
        session_id = request.headers.get("X-Session-ID")
        message_source = request.headers.get("X-Message-Source")

        package_version = sunholo_version()
        tags = [package_version]
        if message_source:
            tags.append(message_source)

        return langfuse.trace(
            name = f"/vac/{vector_name}",
            user_id = user_id,
            session_id = session_id,
            tags = tags,
            release = f"sunholo-v{package_version}"
        )

    def prep_vac(self, request, vector_name):
        trace = self.create_langfuse_trace(request, vector_name)
        span = None

        if request.content_type.startswith('application/json'):
            data = request.get_json()
        elif request.content_type.startswith('multipart/form-data'):
            data = request.form.to_dict()
            if 'file' in request.files:
                file = request.files['file']
                if file.filename != '':
                    log.info(f"Found file: {file.filename} to upload to GCS")
                    try:
                        image_uri, mime_type = self.handle_file_upload(file, vector_name)
                        data["image_uri"] = image_uri
                        data["mime"] = mime_type
                    except Exception as e:
                        return jsonify({'error': str(e), 'traceback': traceback.format_exc()}), 500
                else:
                    return jsonify({"error": "No file selected"}), 400
        else:
            return jsonify({"error": "Unsupported content type"}), 400

        log.info(f"vac/{vector_name} got data: {data}")

        config, _ = load_config("config/llm_config.yaml")
        vac_configs = config.get("vac")
        if vac_configs:
            vac_config = vac_configs[vector_name]

        if trace:
            trace.update(input=data, metadata=vac_config)

        user_input = data.pop('user_input').strip()
        stream_wait_time = data.pop('stream_wait_time', 7)
        stream_timeout = data.pop('stream_timeout', 120)
        chat_history = data.pop('chat_history', None)
        vector_name = data.pop('vector_name', vector_name)

        paired_messages = extract_chat_history(chat_history)

        all_input = {'user_input': user_input, 
                     'vector_name': vector_name, 
                     'chat_history': paired_messages, 
                     'stream_wait_time': stream_wait_time,
                     'stream_timeout': stream_timeout,
                     'kwargs': data}

        if trace:
            span = trace.span(
                name="VAC",
                metadata=vac_config,
                input = all_input
            )
        command_response = handle_special_commands(user_input, vector_name, paired_messages)
        if command_response is not None:
            if trace:
                trace.update(output=jsonify(command_response))
    
        return {
            "trace": trace,
            "span": span,
            "command_response": command_response,
            "all_input": all_input,
            "vac_config": vac_config
        }


    def handle_file_upload(self, file, vector_name):
        try:
            file.save(file.filename)
            image_uri = add_file_to_gcs(file.filename, vector_name)
            os.remove(file.filename)  # Clean up the saved file
            return image_uri, file.mimetype
        except Exception as e:
            raise Exception(f'File upload failed: {str(e)}')





import pytest
from unittest.mock import patch, mock_open
from sunholo.utils import config

def test_load_config():
    expected_config = {"key": "value"}
    with pytest.raises(FileNotFoundError):
        config.load_config("non_existent_file")
    with patch("builtins.open", mock_open(read_data='{"key": "value"}'), create=True):
        result, _ = config.load_config("mock_file.json")
        assert result == expected_config

# Test cases for load_config_key function
@patch("sunholo.utils.config.load_all_configs")
def test_load_config_key(mock_load_all_configs):
    mock_load_all_configs.return_value = {
        "vacConfig": {
            "vac": {
                "test_vector": {
                    "key1": "value1",
                    "key2": "value2"
                }
            }
        }
    }
    # Test existing key
    assert config.load_config_key("key1", "test_vector", "vacConfig") == "value1"
    # Test non-existing key
    result = config.load_config_key("non_existing_key", "test_vector", "vacConfig")
    assert result is None

    # Test invalid configuration
    with pytest.raises(KeyError):
        config.load_config_key("key1", "test_vector", "invalidConfig")



#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import base64
from ..custom_logging import log
from ..gcs.metadata import get_object_metadata



def process_pubsub_message(data: dict) -> tuple:
    """Extracts message data and metadata from a Pub/Sub message.

    Args:
        data (dict): The Pub/Sub message data.

    Returns:
        tuple: A tuple containing the message data and attributes as metadata.
    """
    # Decode the message data
    message_data = base64.b64decode(data['message']['data']).decode('utf-8')
    attributes = data['message'].get('attributes', {})
    messageId = data['message'].get('messageId')
    publishTime = data['message'].get('publishTime')
    vector_name = attributes.get('namespace', None)
    if vector_name:
        attributes['vector_name'] = vector_name

    log.info(f"Process Pub/Sub was triggered by messageId {messageId} published at {publishTime}")
    log.debug(f"Process Pub/Sub data: {message_data}")

    # Check for a valid GCS event type and payload format
    if attributes.get("eventType") == "OBJECT_FINALIZE" and attributes.get("payloadFormat") == "JSON_API_V1":
        objectId = attributes.get("objectId")
        log.info(f"Got valid event from Google Cloud Storage: {objectId}")

        # Ignore config files
        if objectId.startswith("config"):
            log.info("Ignoring config file")
            return None, None, None

        # Construct the message_data
        message_data = 'gs://' + attributes.get("bucketId") + '/' + objectId
        
        if '/' in objectId:
            bucket_vector_name = objectId.split('/')[0]
            if len(bucket_vector_name) > 0 and vector_name != bucket_vector_name:
                log.info(f"Overwriting vector_name {vector_name} with {bucket_vector_name}")
                vector_name = bucket_vector_name
        
        # get metadata for object
        metadata = get_object_metadata(attributes.get("bucketId"), attributes.get("objectId"))
        if metadata:
            metadata.update(attributes)
            attributes = metadata

    return message_data, attributes, vector_name


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..custom_logging import log
from ..utils import ConfigManager
from ..auth import get_header
import requests
import aiohttp
from .langserve import prepare_request_data

from .route import route_endpoint

def prep_request_payload(user_input, chat_history, vector_name, stream, **kwargs):
    """
    Prepares the request payload for sending a query to the QA system.

    Args:
        user_input (str): The user's input query.
        chat_history (list): The history of previous chat messages.
        vector_name (str): The name of the vector.
        stream (bool): Whether the request will be streamed.
        **kwargs: Additional key-value pairs for request data.

    Returns:
        tuple: The endpoint URL and the prepared data payload.

    Example:
    ```python
    user_input = "What is AI?"
    chat_history = []
    vector_name = "my_vector"
    stream = False
    endpoint, payload = prep_request_payload(user_input, chat_history, vector_name, stream)
    print(endpoint, payload)
    # Output: 'http://example.com/invoke' {'user_input': 'What is AI?', 'chat_history': []}
    ```
    """

    config = ConfigManager(vector_name)

    # Add chat_history/vector_name to kwargs so langserve can use them too
    kwargs['chat_history'] = chat_history

    agent = config.vacConfig("agent")
    agent_type = config.vacConfig("agent_type")

    override_endpoint = kwargs.get("override_endpoint")
    if override_endpoint:
        log.info(f"Overriding endpoint with {override_endpoint}")

    # {'stream': '', 'invoke': ''}
    post_endpoints = route_endpoint(override_endpoint=override_endpoint, config=config)

    if stream:
        qna_endpoint = post_endpoints["stream"]
    else: 
        qna_endpoint = post_endpoints["invoke"]

    if agent == "langserve" or agent_type == "langserve":
        get_endpoints = route_endpoint(override_endpoint=override_endpoint, method = 'get', config=config)
        qna_data = prepare_request_data(user_input, get_endpoints["input_schema"], vector_name, **kwargs)
    else:
        # Base qna_data dictionary
        qna_data = {
            'user_input': user_input,
        }
        # Update qna_data with optional values from kwargs
        qna_data.update(kwargs)

        if 'vector_name' not in qna_data:
            qna_data['vector_name'] = vector_name

    return qna_endpoint, qna_data

def send_to_qa(user_input, vector_name, chat_history, stream=False, **kwargs):
    """
    Sends a query to the QA system synchronously.

    Args:
        user_input (str): The user's input query.
        vector_name (str): The name of the vector.
        chat_history (list): The history of previous chat messages.
        stream (bool): Whether to stream the response.
        **kwargs: Additional key-value pairs for request data.

    Returns:
        Union[dict, generator]: The response from the QA system, either as a JSON object or as a generator for streaming.

    Example:
    ```python
    response = send_to_qa("What is AI?", "my_vector", [])
    print(response)
    # Output: {'answer': 'The definition of AI is ...'}
    ```
    """
    qna_endpoint, qna_data = prep_request_payload(user_input, chat_history, vector_name, stream, **kwargs)
    header = get_header(vector_name)
    header = add_header_ids(header, **kwargs)

    log.info(f"Send_to_qa to {qna_endpoint} this data: {qna_data} with this header: {header}")
    try:
        qna_response = requests.post(qna_endpoint, json=qna_data, stream=stream, headers=header)
        qna_response.raise_for_status()

        if stream:
            # If streaming, return a generator that yields response content chunks
            def content_generator():
                for chunk in qna_response.iter_content(chunk_size=8192):
                    yield chunk
            return content_generator()
        else:
            # Otherwise, return the JSON response directly
            return qna_response.json()

    except requests.exceptions.HTTPError as err:
        log.error(f"HTTP error occurred: {err}")
        error_message = f"There was an error processing your request. {str(err)}"
        if stream:
            return iter([error_message])
        else:
            return {"answer": error_message}

    except Exception as err:
        log.error(f"Other error occurred: {str(err)}")
        error_message = f"Something went wrong. {str(err)}"
        if stream:
            return iter([error_message])
        else:
            return {"answer": error_message}

async def send_to_qa_async(user_input, vector_name, chat_history, stream=False, **kwargs):
    """
    Sends a query to the QA system asynchronously.

    Args:
        user_input (str): The user's input query.
        vector_name (str): The name of the vector.
        chat_history (list): The history of previous chat messages.
        stream (bool): Whether to stream the response.
        **kwargs: Additional key-value pairs for request data.

    Returns:
        generator: A generator that yields chunks of the streamed response or the entire response.

    Example:
    ```python
    async def example_async_query():
        async for response_chunk in send_to_qa_async("What is AI?", "my_vector", []):
            print(response_chunk)
    ```
    """    
    qna_endpoint, qna_data = prep_request_payload(user_input, chat_history, vector_name, stream, **kwargs)
    header = get_header(vector_name)
    header = add_header_ids(header, **kwargs)

    log.info(f"send_to_qa_async to {qna_endpoint} this data: {qna_data} with this header: {header}")

    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(qna_endpoint, json=qna_data, headers=header) as resp:
                resp.raise_for_status()

                if stream:
                    # Stream the response
                    async for chunk in resp.content.iter_any():
                        yield chunk
                else:
                    # Return the complete response
                    qna_response = await resp.json()
                    log.info(f"Got back QA response: {qna_response}")
                    yield qna_response
    except aiohttp.ClientResponseError as e:
        log.error(f"HTTP error occurred: {e}")
        error_message = f"There was an error processing your request: {str(e)}"
        if stream:
            yield error_message.encode('utf-8')
        else:
            yield {"answer": error_message}
    except Exception as e:
        log.error(f"Other error occurred: {str(e)}")
        error_message = f"Something went wrong: {str(e)}"
        if stream:
            yield error_message.encode('utf-8')
        else:
            yield {"answer": error_message}


def add_header_ids(header, **kwargs):
    """
    Adds user and session IDs to the request header if they are provided.

    Args:
        header (dict): The original request header.
        **kwargs: Additional key-value pairs for header data.

    Returns:
        dict: The updated header.

    Example:
    ```python
    header = {"Authorization": "Bearer some_token"}
    updated_header = add_header_ids(header, user_id="user_123", session_id="session_456")
    print(updated_header)
    # Output: {'Authorization': 'Bearer some_token', 'X-User-ID': 'user_123', 'X-Session-ID': 'session_456'}
    ```
    """
    if not header:
        return None
    
    # Check if 'user_id' and 'session_id' are in kwargs and add them to the header
    if 'user_id' in kwargs:
        header['X-User-ID'] = kwargs['user_id']  # Custom header names, adjust as needed
    if 'session_id' in kwargs:
        header['X-Session-ID'] = kwargs['session_id']
    if 'message_source' in kwargs:
        header['X-Message-Source'] = kwargs['message_source']
    
    return header


from ..custom_logging import log
from ..utils.config import load_config_key
from ..utils.gcp_project import get_gcp_project
from ..components import load_memories

from .discovery_engine_client import DiscoveryEngineClient
from .create_new import create_new_discovery_engine
    

def do_discovery_engine(message_data, metadata, vector_name):
    """

    Example:
    ```python
    message_data = "gs://bucket_name/path_to_file.txt"
    metadata = {"user": "admin"}
    vector_name = "example_vector"
    response = do_discovery_engine(message_data, metadata, vector_name)
    print(response)
    # Imported file to corpus: {'status': 'success'}
    ```
    """

    memories = load_memories(vector_name)
    tools = []

    if not memories:
        return tools
    
    corpuses = []
    for memory in memories:
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore == "discovery_engine" or vectorstore == "vertex_ai_search":
                log.info(f"Found vectorstore {vectorstore}")
                if value.get('read_only'):
                    continue
                #location = gcp_config.get('location') 
                corpus = DiscoveryEngineClient(
                    data_store_id=vector_name, 
                    project_id=get_gcp_project(),
                    # location needs to be 'eu' or 'us' which doesn't work with other configurations
                    #location=location or global_location
                    )

                corpuses.append(corpus)
    if not corpuses:
        log.error("Could not find any Discovery Engine corpus to import data to")
        return None

    log.info(f"Found Discovery Engine / Vertex AI Search {corpuses=}")

    if message_data.startswith("gs://"):
        log.info(f"DiscoveryEngineClient.import_files for {message_data}")
        if "/pdf_parts/" in message_data:
            return None
        for corp in corpuses:
            try:
                response = corp.import_documents(
                    gcs_uri=message_data
                )
                log.info(f"Imported file to corpus: {response} with metadata: {metadata}")
            except Exception as err:
                log.error(f"Error importing {message_data} - {corp=} - {str(err)}")

                if str(err).startswith("404"):
                    log.info(f"Attempting to create a new DiscoveryEngine corpus: {vector_name}")
                    try:
                        new_corp = create_new_discovery_engine(vector_name)
                    except Exception as err:
                        log.error(f"Failed to create new DiscoveryEngine {vector_name} - {str(err)}")
                        continue
                    if new_corp:
                        log.info(f"Found new DiscoveryEngine {vector_name=} - {new_corp=}")
                        response = corp.import_documents(
                            gcs_uri=message_data
                        )
                    
                continue

        metadata["source"] = message_data
        return metadata
    
    else:
        log.warning("Only gs:// data is supported for Discovery Engine")


def check_discovery_engine_in_memory(vector_name):
    memories = load_config_key("memory", vector_name=vector_name, kind="vacConfig")
    for memory in memories:  # Iterate over the list
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore:
                if vectorstore == "discovery_engine" or vectorstore == "vertex_ai_search":
                    log.info(f"Found vectorstore {vectorstore}")
                    return True
    
    return False

def discovery_engine_chunker_check(message_data, metadata, vector_name):
    # discovery engine handles its own chunking/embedding
    memories = load_config_key("memory", vector_name=vector_name, kind="vacConfig")
    if not memories:
        return None
    
    total_memories = len(memories)
    llama = None
    if check_discovery_engine_in_memory(vector_name):
        llama = do_discovery_engine(message_data, metadata, vector_name)
        log.info(f"Processed discovery engine: {llama}")

    # If discovery engine is the only entry, return
    if llama and total_memories == 1:

        return llama
    
    elif llama:
        log.info("Discovery Engine found but not the only memory, continuing with other processes.")

        return None


import os
import tempfile

try:
    from vertexai.preview import rag
except ImportError:
    rag = None

from ..utils import ConfigManager
from ..custom_logging import log

try:
    from ..cli.sun_rich import console
except ImportError:
    console = None

class LlamaIndexVertexCorpusManager:
    """
    A manager class for handling RAG corpora operations with Vertex AI.

    Attributes:
        config (ConfigManager): Configuration manager for fetching project settings.
        project_id (str): GCP project ID.
        location (str): GCP location.
    """

    def __init__(self, config: ConfigManager = None, project_id: str = None, location: str = None):
        """
        Initializes the LlamaIndexVertexCorpusManager.

        Args:
            config (ConfigManager): Configuration manager for fetching project settings.
            project_id (str): GCP project ID.
            location (str): GCP location.

        Raises:
            ImportError: If vertexai is not installed.
        """
        from ..vertex.init import init_vertex
        if rag is None:
            raise ImportError("You must install vertexai via `pip install sunholo[gcp]`")
        
        self.config = config
        self.project_id = project_id
        self.location = location

        if config:
            self.project_id = self.config.vacConfig('project_id') or project_id
            self.location = self.config.vacConfig('location') or location

        init_vertex(location=self.location, project_id=self.project_id)

    def upload_text(self, text: str, corpus_display_name: str, description: str = None):
        """
        Uploads a text string to a specified corpus by saving it to a temporary file first.

        Args:
            text (str): The text content to upload.
            corpus_display_name (str): The display name of the corpus.
            description (str, optional): Description of the text upload.

        Returns:
            The uploaded file object.
        """
        # Create a temporary file and write the text to it
        with tempfile.NamedTemporaryFile(delete=False, suffix='.txt') as temp_file:
            temp_file.write(text.encode('utf-8'))
            temp_filename = temp_file.name
        
        try:
            # Upload the temporary file
            uploaded_file = self.upload_file(temp_filename, corpus_display_name, description or text[:50])
        finally:
            # Clean up the temporary file
            os.remove(temp_filename)

        return uploaded_file
    
    def upload_file(self, filename:str, corpus_display_name:str, description:str=None):
        corpus = self.find_corpus_from_list(corpus_display_name)

        rag_file = rag.upload_file(
            corpus_name=corpus.name,
            path=filename,
            display_name=filename,
            description=description or f"Upload for {filename}",
        )
        log.info("Uploaded file: {rag_file}")

        return rag_file
    
    def import_files(self, file_imports:list[str], corpus_display_name:str):
        corpus = self.find_corpus_from_list(corpus_display_name)

        log.info(f"Importing files: {file_imports} into {corpus.name}")
        response = rag.import_files(
            corpus_name=corpus.name,
            paths=file_imports,
            chunk_size=512,  # Optional
            chunk_overlap=100,  # Optional
            max_embedding_requests_per_min=900,  # Optional
        )
        log.info(f"Imported files: {response}")

        return response
    
    def list_corpora(self):
        """
        List all VertexAI Corpus for the project/location
        """
        return rag.list_corpora()
    
    def find_corpus_from_list(self, display_name: str):
        """
        Finds a corpus from the list of corpora by its display name.

        Args:
            display_name (str): The display name of the corpus.

        Returns:
            The found corpus object if it exists, otherwise None.
        """
        corpora = self.list_corpora()
        for corp in corpora:
            if display_name == corp.display_name:
                log.info(f"Found existing corpus with display name: {display_name}")
                return corp
        return None

    def create_corpus(self, display_name: str, description: str = None):
        """
        Creates a new corpus or returns an existing one with the specified display name.

        Args:
            display_name (str): The display name of the corpus.
            description (str, optional): Description of the corpus.

        Returns:
            The created or found corpus object.
        """
        corp = self.find_corpus_from_list(display_name)
        if corp:
            return corp
        
        corpus = rag.create_corpus(display_name=display_name, description=description or f"Corpus for {display_name}")
        log.info(f"created Llamaindex corpus {corpus}")

        return corpus

    def delete_corpus(self, display_name: str):
        """
        Deletes a corpus by its display name.

        Args:
            display_name (str): The display name of the corpus.

        Returns:
            bool: True if the corpus was deleted, False otherwise.
        """
        corp = self.find_corpus_from_list(display_name)
        if corp:
            rag_id = corp.name
            rag.delete_corpus(name=rag_id)
            log.info(f"Deleted {rag_id}")
            return True
        
        log.warning(f"Could not find a corp to delete with name {display_name}")
        return False

    def fetch_corpus(self, display_name: str):
        """
        Fetches a corpus by its display name.

        Args:
            display_name (str): The display name of the corpus.

        Returns:
            The fetched corpus object.

        Raises:
            ValueError: If the corpus with the specified display name does not exist.
        """
        corp = self.find_corpus_from_list(display_name)
        if not corp:
            raise ValueError(f"Could not find any corpus with display_name: {display_name}")
        return rag.get_corpus(name=corp.name)

def llamaindex_command(args):
    if console is None:
        raise ImportError("Need cli tools to use `sunholo llamaindex` - install via `pip install sunholo[cli]`")
    
    config = ConfigManager(args.vac)
    manager = LlamaIndexVertexCorpusManager(config=config)
    
    if args.action == "create":
        manager.create_corpus(display_name=args.display_name, description=args.description)
    elif args.action == "delete":
        manager.delete_corpus(display_name=args.display_name)
    elif args.action == "fetch":
        corpus = manager.fetch_corpus(display_name=args.display_name)
        console.print(corpus)
    elif args.action == "find":
        corpus = manager.find_corpus_from_list(display_name=args.display_name)
        console.print(corpus)
    elif args.action == "list":
        corpora = manager.list_corpora()
        console.print(corpora)
    elif args.action == "import_files":
        manager.import_files(file_imports=args.file_imports, corpus_display_name=args.display_name)
    elif args.action == "upload_file":
        manager.upload_file(filename=args.filename, corpus_display_name=args.display_name, description=args.description)
    elif args.action == "upload_text":
        manager.upload_text(text=args.text, corpus_display_name=args.display_name, description=args.description)
    else:
        console.print(f"Unknown action: {args.action}")

def setup_llamaindex_subparser(subparsers):
    """
    Sets up an argparse subparser for the 'llamaindex' command.

    Args:
        subparsers: The subparsers object from argparse.ArgumentParser().
    """
    # LlamaIndex subparser setup
    llamaindex_parser = subparsers.add_parser('llamaindex', help='Manage LlamaIndex corpora')
    llamaindex_subparsers = llamaindex_parser.add_subparsers(dest='action', help='LlamaIndex subcommands')

    # LlamaIndex create command
    create_parser = llamaindex_subparsers.add_parser('create', help='Create a new corpus')
    create_parser.add_argument('display_name', help='The name of the corpus')
    create_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')
    create_parser.add_argument('--description', help='Description of the corpus', default=None)

    # LlamaIndex delete command
    delete_parser = llamaindex_subparsers.add_parser('delete', help='Delete a corpus')
    delete_parser.add_argument('display_name', help='The name of the corpus')
    delete_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')

    # LlamaIndex fetch command
    fetch_parser = llamaindex_subparsers.add_parser('fetch', help='Fetch a corpus')
    fetch_parser.add_argument('display_name', help='The name of the corpus')
    fetch_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')

    # LlamaIndex find command
    find_parser = llamaindex_subparsers.add_parser('find', help='Find a corpus')
    find_parser.add_argument('display_name', help='The name of the corpus')
    find_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')

    # LlamaIndex list command
    list_parser = llamaindex_subparsers.add_parser('list', help='List all corpus')
    list_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')

    # LlamaIndex import_files command
    import_files_parser = llamaindex_subparsers.add_parser('import_files', help='Import files from URLs to a corpus')
    import_files_parser.add_argument('display_name', help='The name of the corpus')
    import_files_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')
    import_files_parser.add_argument('file_imports', nargs='+', help='The list of file URLs to import')

    # LlamaIndex upload_file command
    upload_file_parser = llamaindex_subparsers.add_parser('upload_file', help='Upload a local file to a corpus')
    upload_file_parser.add_argument('display_name', help='The name of the corpus')
    upload_file_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')
    upload_file_parser.add_argument('filename', help='The local file path to upload')
    upload_file_parser.add_argument('--description', help='Description of the file upload', default=None)

    # LlamaIndex upload_text command
    upload_text_parser = llamaindex_subparsers.add_parser('upload_text', help='Upload text to a corpus')
    upload_text_parser.add_argument('display_name', help='The name of the corpus')
    upload_text_parser.add_argument('vac', nargs='?', default="global", help='The VAC config to set it up for')
    upload_text_parser.add_argument('text', help='The text content to upload')
    upload_text_parser.add_argument('--description', help='Description of the text upload', default=None)

    llamaindex_parser.set_defaults(func=llamaindex_command)
    


import os
import json
import yaml
from datetime import datetime, timedelta
from collections import defaultdict
from .timedelta import format_timedelta

class ConfigManager:
    def __init__(self, vector_name: str, validate:bool=True):
        """
        Initialize the ConfigManager with a vector name.
        Requires a local config/ folder holding your configuration files or the env var VAC_CONFIG_FOLDER to be set.

        Read more at: https://dev.sunholo.com/docs/config

        Args:
            vector_name (str): The name of the vector in the configuration files.
            validate (bool): Whether to validate the configurations
        
        Example:
        ```python
        # Usage example:
        config = ConfigManager("my_vac")
        agent = config.vacConfig("agent")
        ```
        """
        local_config_folder = os.path.join(os.getcwd(), "config")
        if os.path.isdir(local_config_folder):
            print(f"Found local config folder {local_config_folder} - will overwrite any global configurations")
        else:
            local_config_folder = None
        if os.getenv("VAC_CONFIG_FOLDER") is None and local_config_folder is None:
            raise ValueError(f"Must have either a local config/ folder in this dir ({os.getcwd()}/config/) or a folder specified via the VAC_CONFIG_FOLDER environment variable, or both.")

        self.vector_name = vector_name
        self.config_cache = {}
        self.config_folder = os.getenv("VAC_CONFIG_FOLDER", os.getcwd())
        self.local_config_folder = local_config_folder
        self.configs_by_kind = self.load_all_configs()

        test_agent = self.vacConfig("agent")
        if not test_agent and self.vector_name != "global" and validate:
            print(f"WARNING: No vacConfig.agent found for {self.vector_name} - are you in right folder? {local_config_folder=} {self.config_folder=}")

    def load_all_configs(self):
        """
        Load all configuration files from the specified directories into a dictionary.
        Caching is used to avoid reloading files within a 5-minute window.

        Returns:
            dict: A dictionary of configurations grouped by their 'kind' key.
        """
        from ..custom_logging import log

        log.debug(f"Loading all configs from folder: {self.config_folder} and local folder: {self.local_config_folder}")
        global_configs_by_kind = self._load_configs_from_folder(self.config_folder)

        if self.local_config_folder:
            local_configs_by_kind = self._load_configs_from_folder(self.local_config_folder)
            # Merge local configs into global configs
            for kind, local_config in local_configs_by_kind.items():
                if kind in global_configs_by_kind:
                    global_configs_by_kind[kind] = self._merge_dicts(global_configs_by_kind[kind], local_config)
                else:
                    global_configs_by_kind[kind] = local_config

        return global_configs_by_kind

    def _load_configs_from_folder(self, folder):
        """
        Load all configuration files from a specific folder into a dictionary.

        Args:
            folder (str): The path of the folder to load configurations from.

        Returns:
            dict: A dictionary of configurations grouped by their 'kind' key.
        """
        from ..custom_logging import log

        configs_by_kind = defaultdict(dict)
        current_time = datetime.now()
        
        for filename in os.listdir(folder):
            if filename in ["cloudbuild.yaml", "cloud_run_urls.json"]:
                continue
            if filename.endswith(('.yaml', '.yml', '.json')):
                config_file = os.path.join(folder, filename)
                if filename in self.config_cache:
                    cached_config, cache_time = self.config_cache[filename]
                    time_to_recache = (current_time - cache_time)
                    if time_to_recache < timedelta(minutes=5):
                        config = cached_config
                    else:
                        config = self._reload_config_file(config_file, filename, folder == self.local_config_folder)
                else:
                    config = self._reload_config_file(config_file, filename, folder == self.local_config_folder)
                if config is None:
                    log.error(f"No config found within {filename}")
                    continue
                kind = config.get('kind')
                if kind:
                    configs_by_kind[kind] = config
                else:
                    log.warning(f"No 'kind' found in {filename}")
        return configs_by_kind

    def _reload_config_file(self, config_file, filename, is_local=False):
        """
        Helper function to load a config file and update the cache.

        Args:
            config_file (str): The path to the configuration file.
            filename (str): The name of the configuration file.
            is_local (bool): Indicates if the config file is from the local folder.

        Returns:
            dict: The loaded configuration.
        """
        from ..custom_logging import log
        with open(config_file, 'r') as file:
            if filename.endswith('.json'):
                config = json.load(file)
            else:
                config = yaml.safe_load(file)
        self.config_cache[filename] = (config, datetime.now())
        log.debug(f"Loaded and cached {config_file}")
        if is_local:
            log.warning(f"Local configuration override for {filename} via {self.local_config_folder}")
        return config

    def _check_and_reload_configs(self):
        """
        Check if configurations are older than 5 minutes and reload if necessary.
        """
        current_time = datetime.now()
        for filename, (config, cache_time) in list(self.config_cache.items()):
            if (current_time - cache_time) >= timedelta(minutes=5):
                config_file_main = os.path.join(self.config_folder, filename)
                config_file_local = os.path.join(self.local_config_folder, filename)
                if os.path.exists(config_file_local):
                    self._reload_config_file(config_file_local, filename, is_local=True)
                if os.path.exists(config_file_main):
                    self._reload_config_file(config_file_main, filename, is_local=False)
        self.configs_by_kind = self.load_all_configs()

    def _merge_dicts(self, dict1, dict2):
        """
        Recursively merge two dictionaries. Local values in dict2 will overwrite global values in dict1.

        Args:
            dict1 (dict): The global dictionary.
            dict2 (dict): The local dictionary.

        Returns:
            dict: The merged dictionary.
        """
        for key, value in dict2.items():
            if isinstance(value, dict) and key in dict1 and isinstance(dict1[key], dict):
                dict1[key] = self._merge_dicts(dict1[key], value)
            else:
                dict1[key] = value
        return dict1

    def vacConfig(self, key: str):
        """
        Fetch a key from 'vacConfig' kind configuration.

        Args:
            key (str): The key to fetch from the configuration.

        Returns:
            str: The value associated with the specified key.
        """
        self._check_and_reload_configs()
        config = self.configs_by_kind.get('vacConfig')
        if not config:
            return None
        if self.vector_name == 'global':
            return config.get(key)
        vac = config['vac']

        vac_config = vac.get(self.vector_name)
        if not vac_config:
            return None
        return vac_config.get(key)

    def promptConfig(self, key: str):
        """
        Fetch a key from 'promptConfig' kind configuration.

        Args:
            key (str): The key to fetch from the configuration.

        Returns:
            str: The value associated with the specified key.
        """
        self._check_and_reload_configs()
        config = self.configs_by_kind.get('promptConfig')
        if not config:
            return None
        prompts = config['prompts']
        prompt_for_vector_name = prompts.get(self.vector_name)
        if not prompt_for_vector_name:
            return None
        return prompt_for_vector_name.get(key)

    def agentConfig(self, key: str):
        """
        Fetch a key from 'agentConfig' kind configuration.

        Args:
            key (str): The key to fetch from the configuration.

        Returns:
            str: The value associated with the specified key.
        """
        self._check_and_reload_configs()
        config = self.configs_by_kind.get('agentConfig')
        if not config:
            return None
        agents = config.get('agents')
        if key in agents:
            return agents[key]
        else:
            return agents.get("default")

